;;; doubly-linked-list.el --- -*- coding: iso-8859-1 -*-
;; Copyright 2012 Vincent Belache
;;
;; Author: Vincent Belache <vincentb1@users.sourceforge.net>
;; Version: $Id: doubly-linked-list.el,v 1.4 2012-11-29 20:29:00 Vincent Exp $
;; Keywords:
;; X-URL: not distributed yet

;;
;; This software is governed by the CeCILL license under French law and abiding
;; by the rules of distribution of free software.  You can use, modify and/ or
;; redistribute the software under the terms of the CeCILL license as circulated
;; by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
;;
;; As a counterpart to the access to the source code and rights to copy, modify
;; and redistribute granted by the license, users are provided only with a
;; limited warranty and the software's author, the holder of the economic
;; rights, and the successive licensors have only limited liability.
;;
;; In this respect, the user's attention is drawn to the risks associated with
;; loading, using, modifying and/or developing or reproducing the software by
;; the user in light of its specific status of free software, that may mean that
;; it is complicated to manipulate, and that also therefore means that it is
;; reserved for developers and experienced professionals having in-depth
;; computer knowledge. Users are therefore encouraged to load and test the
;; software's suitability as regards their requirements in conditions enabling
;; the security of their systems and/or data to be ensured and, more generally,
;; to use and operate it in the same conditions as regards security.
;;
;; The fact that you are presently reading this means that you have had
;; knowledge of the CeCILL license and that you accept its terms.
;;
;;; Commentary:

;;

;; Put this file into your load-path and the following into your ~/.emacs:
;;   (require 'doubly-linked-list)

;;; Code:
(require 'eieio)
(eval-when-compile
  (require 'cl))



;;;;##########################################################################
;;;;  User Options, Variables
;;;;##########################################################################

(defclass doli2-list ()
  ((head
    :initarg :head
    :initform nil
    :documentation "Pointe sur le premier lment de la liste si
   la liste est non-vide.")
   (tail
    :initarg :tail
    :initform nil
    :documentation "Pointe sur le dernier lment de la liste si
    la liste est non vide.")
   (size
    :initarg :size
    :initform 0
    :documentation "Nombre d'lments de la liste."))
  :documentation "Liste doublement chane.")

(defclass doli2-element ()
  ((next
    :initarg :next
    :initform nil
    :type (or null doli2-element)
    :documentation "lment suivant dans la liste, `nil' pour le
    dernier lment.")
   (previous
    :initarg :previous
    :initform nil
    :type (or null doli2-element)
    :documentation "lment prcdent dans la liste, `nil'
    pour le dernier lment."  ))
  :abstract t
  :documention "Classe de base pour les lments d'objets de classe
  `doli2-list'.")

(defgeneric doli2-equal ((this doli2-element) (other doli2-element))
  "Comparaison de deux `doli2-element'.")

(defmethod doli2-add-first
  ((this doli2-list) elt
   ;(elt doli2-element )
   )
  "Ajoute un nouvel lment en tte de liste."
  (when
      (or (oref elt :next)
	  (oref elt :previous))
    (error "L'lment est dj dans une liste"))
  (let ((size (oref this :size)))
    (if (= 0 size)
	(oset this :tail elt)
      (let ((head  (oref this :head)))
	(oset elt :next head)
	(oset head :previous elt)))
    (oset this :head elt)
    (oset this :size (1+ size))))

(defmethod doli2-add-last
  ((this doli2-list) elt
   ;(elt doli2-element)
   )
  "Ajoute un nouvel lment en queue de liste."
  (when
      (or (oref elt :next)
	  (oref elt :previous))
    (error "L'lment est dj dans une liste"))
  (let ((size (oref this :size)))
    (if (= 0 size)
	  (oset this :head elt)
      (let ((tail  (oref this :tail)))
	(oset elt :previous tail)
	(oset tail :next elt)))
    (oset this :tail elt)
    (oset this :size (1+ size))))

(defmethod doli2-remove
  ((this doli2-list)
   elt;(elt doli2-element)
   )
  "Retranche `elt' de la liste `this'."
  (let ((size (oref this :size)))
    (cond
     ((>= 0 size)
      (error "Liste vide %S" this))
     ((= 1 size)
      (oset this :head nil)
      (oset this :tail nil)
      (oset this :size 0))
     ((eq (oref this :head) elt)
      (let ((next (oref elt :next)))
	(oset this :head next)
	(oset next :previous nil)
	(oset this :size (1- size))
	(oset elt :previous nil)
	(oset elt :next nil)))
     ((eq (oref this :tail) elt)
      (let ((previous (oref elt :previous)))
	(oset this :tail previous)
	(oset previous :next nil)
	(oset this :size (1- size))
	(oset elt :previous nil)
	(oset elt :next nil)))
     (t
      (let ((next (oref elt :next))
	    (previous (oref elt :previous)))
	(oset next :previous previous)
	(oset previous :next next)
	(oset this :size (1- size))
	(oset elt :previous nil)
	(oset elt :next nil))))))

(defmethod doli2-remove-first
  ((this doli2-list))
  "Retranche le premier lment de la liste THIS. Gnre une
erreur si la liste est vide."
  (let ((ret (oref this :head))
	(size (1- (oref this :size))))
    (if (> 0 size)
	(error "Liste vide")
      (oset this :size size)
      (if (< 0  size)
	  (let ((next (oref ret :next)))
	    (oset this :head next)
	    (when next (oset next :previous nil)))
	(oset this :head nil)
	(oset this :tail nil)))
     (oset ret :next nil)
     ret))

(defmethod doli2-remove-last
  ((this doli2-list))
  "Retranche le dernier lment de la liste THIS. Gnre une
erreur si la liste est vide."
  (let ((ret (oref this :tail))
	(size (1- (oref this :size))))
    (if (> 0 size)
	(error "Liste vide")
      (oset this :size size)
      (if (< 0  size)
	 (let ((previous  (oref ret :previous)))
	  (oset this :tail previous)
	  (when previous (oset previous :next nil)))
	(oset this :tail nil)
	(oset this :head nil)))
     (oset ret :previous nil)
     ret))

(defmacro doli2-.>x-dolist (spec &rest body)
  "(doli2-.>x-dolist (V FROM TO) BODY) itre V sur BODY depuis FROM
inclus, jusqu' TO exclus en parcourant la liste dans le sens normal."
  `(let ((,(car spec) ,(cadr spec)))
     (while (null (eq ,(car spec) ,(caddr spec)))
       ,@body
       (setq ,(car spec) (oref ,(car spec) :next)))))

(defmacro doli2-x>x-dolist (spec &rest body)
  "(doli2-x>x-dolist (V FROM TO) BODY) itre V sur BODY depuis FROM
exclus, jusqu' TO exclus en parcourant la liste dans le sens normal."
  `(let ((,(car spec) (oref ,(cadr spec) :next)))
     (unless (eq ,(car spec) ,(caddr spec))
       (while
	   (progn
	     ,@body
	     (setq ,(car spec) (oref ,(car spec) :next))
	     (null (eq ,(car spec) ,(caddr spec))))))))

(defmacro doli2-.>-dolist (spec &rest body)
  "(doli2-.>-dolist (V FROM) BODY) itre sur BODY depuis FROM inclus,
jusqu' la fin inculse de la liste en parcourant la liste dans le
sens normal."
  `(let ((,(car spec) ,(cadr spec)))
     (while
       (progn
	 ,@body
	 (setq ,(car spec) (oref ,(car spec) :next))))))

(defmacro doli2-x>-dolist (spec &rest body)
  "(doli2-x>-dolist (V FROM) BODY) itre sur BODY depuis FROM exclus,
jusqu' la fin inculse de la liste en parcourant la liste dans le
sens normal."
  `(let ((,(car spec) (oref ,(cadr spec) :next)))
     (when ,(car spec)
       (while
	   (progn
	     ,@body
	     (setq ,(car spec) (oref ,(car spec) :next)))))))

(defmacro doli2-x<.-dolist (spec &rest body)
  "(doli2-.>x-dolist (V FROM TO) BODY) itre sur BODY depuis FROM
inclus, jusqu' TO exclus dans le sens inverse."
  `(let ((,(car spec) ,(cadr spec)))
     (while (null (eq ,(car spec) ,(caddr spec)))
      ,@body
      (setq ,(car spec) (oref ,(car spec) :previous)))))


(defmacro doli2-<.-dolist (spec &rest body)
  "(doli2-<.-dolist (V FROM) BODY) itre V sur BODY depuis FROM
inclus, jusqu'au dbut inclus de la liste en parcourant la
liste dans le sens inverse."
  `(while
       (progn
	 ,@body
	 (setq ,spec (oref ,spec :previous)))))

(defmacro doli2->-dolist (spec &rest body)
  "(doli2->-dolist (V LIST) BODY). Itre sur BODY avec V qui
parcourt LIST dans le sens normal."
  `(let ((,(car spec) (oref ,(cadr spec) :head)))
     (while ,(car spec)
       ,@body
       (setq ,(car spec) (oref ,(car spec) :next)))))

(defmacro doli2-<-dolist (spec &rest body)
  "(doli2->-dolist (V LIST) BODY). Itre sur BODY avec V qui
parcourt LIST dans le sens inverse."
  `(let ((,(car spec) (oref ,(cadr spec) :tail)))
     (while ,(car spec)
       ,@body
       (setq ,(car spec) (oref ,(car spec) :previous)))))

(provide 'doubly-linked-list)

;;; doubly-linked-list.el ends here
